"""VirtualHostFolder module

Defines the VirtualHostFolder ("VHF") class

Based, obviously, on VirtualHostMonster--thanks to Evan Simpson and
iMeme.  Comments and suggestions to garyposter@earthlink.net

Requires Python 2.x, Zope 2.4+
"""

# standard issue Zope and Python modules
from Globals import DTMLFile, MessageDialog, InitializeClass
from AccessControl import ClassSecurityInfo
from Acquisition import aq_parent, aq_base
from ZPublisher import BeforeTraverse
import re
# from OFS.FindSupport import FindSupport
from OFS.PropertyManager import PropertyManager
from OFS.SimpleItem import SimpleItem
from OFS.ObjectManager import ObjectManager
from OFS.Folder import Folder

#non-standard Zope requisite:
from Products.OrderedFolder import OrderedObjectManager, SubObjectTypeSupport

# my own
import IVirtualHostFolder, IHostOverride, IHost, utils


class VirtualHostFolder(OrderedObjectManager.OrderedObjectManager, Folder):
    """a folder for, um, virtual hosts.  See the interface."""

    security = ClassSecurityInfo()
    security.declareObjectPublic() # uh...yeah, I think so...

    __implements__=IVirtualHostFolder.IVirtualHostFolder
    # and SimpleItem and other Zope
    # mix-ins 
    
    def possible_meta_types(self):
        return ObjectManager.all_meta_types(self, (IHost.IHost,))

    meta_type="Virtual Host Folder"
    id='VHF' # the thought was to force this...not doing that right now...
    priority = 25
    title = 'Virtual Host Folder'

    manage_options=(
        (
        {'label':'Contents', 'action':'manage_main', 'help':('VirtualHostFolder','VHF_Configure.stx')}, 
        {'label':'Test', 'action':'manage_TestForm',
        'help':('VirtualHostFolder','VHF_TestDomainPath.stx')},
        {'label':'Evaluate', 'action':'manage_evaluateVHF',
         'help':('VirtualHostFolder','VHF_Evaluate.stx')}
        )
        + PropertyManager.manage_options
        + SubObjectTypeSupport.SubObjectTypeSupport.manage_options
        + SimpleItem.manage_options
        # + FindSupport.manage_options
        ) # removed RoleManager.manage_options

    security.declareProtected( 'View management screens', 'manage_evaluateVHF' )
    manage_evaluateVHF = DTMLFile('dtml/VHFevaluate', globals())

    security.declareProtected( 'View management screens', 'manage_TestForm' )
    manage_TestForm = DTMLFile('dtml/VHFtest', globals())

    #####################################################
    
    #security already defined elsewhere(CopySupport, presumably)
    #security.declareProtected( 'View management screens', 'manage_renameItems' )
    def manage_renameObject(self, id, new_id, REQUEST=None):
        """Rename a particular sub-object"""
        result=OrderedObjectManager.OrderedObjectManager.manage_renameObject(id, new_id, REQUEST)
        if not REQUEST and hasattr(self, "REQUEST"):
            REQUEST=self.REQUEST
        if REQUEST and REQUEST.has_key("VIRTUAL_HOST_IDS"):
            currentIdPath=REQUEST["VIRTUAL_HOST_IDS"].split("/")
            if id==currentIdPath[0]:
                currentIdPath[0]=new_id
                REQUEST.set( "VIRTUAL_HOST_IDS", "/".join(currentIdPath) )
        return result

    security.declareProtected( 'View', 'testCall' )
    def testCall(self, URL):
        """A way to test the VHF without actually changing anything"""
        URL=URL.strip()
        res=utils.testURLMatch.match(URL)
        if res:
            protocol, host, port, path=res.group(1,2,3,4)
            fakeStack=[]
            if path and len(path)>1:
                path=path[1:].split("/")
                if not path[-1]:
                    path.pop()
                path.reverse()
                fakeStack.extend(path)
            fakeStack.append(host+(port or ""))
            fakeStack.append(protocol and protocol[:-3] or "http") # :-3 gets rid of :// from, for instance, http://
            fakeStack.append('VirtualHostBase')
            fakeRequest=utils.RequestMimic()
            fakeRequest['TraversalRequestNameStack']=fakeStack
            fakeRequest['HTTP_HOST']=host+(port and port[1:] or "80")
            fakeRequest['BASE0']=(protocol and protocol[:-3] or "http")+"://"+host
            try:
                self(None, fakeRequest)
            except "Redirect", e:
                fakeRequest['FAKE_REDIRECT']=e
            return fakeRequest
        return None

    def __call__(self, client, request, response=None):
        '''Traversing at home'''
        stack = request['TraversalRequestNameStack']
        host = ''
        if stack and stack[-1] == 'VirtualHostBase':
            stack.pop()
            protocol = stack.pop()
            host = stack.pop()
            if ':' in host:
                host, port = host.split(':')
                request.setServerURL(protocol, host, port)
            else:
                request.setServerURL(protocol, host)
        if host == '':
           host = request['HTTP_HOST'].split(':')[0]
        # If a domain to path mapping applies, invoke it now.
        if not stack.count('VirtualHostRoot'):
            breadCrumbs=""
            for loc in stack:
                if loc: breadCrumbs="/"+loc+breadCrumbs
            request.set ("VIRTUAL_HOST_BREADCRUMBS", breadCrumbs)
            myId=self.getId()
            for hostObj in self.objectValues():
                res= hostObj.testActivate(host, stack, myId, request)
                if res:
                    hostObj.activate(res, host, stack, myId, request)
                    break
        else:####### BLACK HOLE FOLLOWS!! ############
            # below untouched other than placing in else clause,
            # switching to string methods, replacing .id with
            # .getId().  It may not even work anymore--certainly many
            # of the convenience features I've created will have
            # problems.

            # Find and convert VirtualHostRoot directive
            # If it is followed by one or more path elements that each
            # start with '_vh_', use them to construct the path to the
            # virtual root.
            vh = -1
            for ii in range(len(stack)):
                if stack[ii] == 'VirtualHostRoot':
                    if vh >= 0:
                        pp = ['']
                        for jj in range(vh, ii):
                            pp.insert(1, stack[jj][4:])
                        stack[vh:ii + 1] = [pp.join('/'), self.getId()]
                    elif ii > 0 and stack[ii - 1][:1] == '/':
                        stack[ii] = self.getId()
                    else:
                        stack[ii] = self.getId()
                        stack.insert(ii, '/')
                    break
                elif vh < 0 and stack[ii][:4] == '_vh_':
                    vh = ii

    def __bobo_traverse__(self, request, name):
        '''Traversing away'''
        # we are doing normal traversal if the name isn't (or doesn't begin
        # with) a forward slash.
        if not name.startswith("/"): # [:1] != '/':
            return self._getOb(name, None)
        # else we are dealing with the part of the path that __call__ set up (or
        # that the manager hardcoded in the external server or included as
        # testing directly in the URL, via old-style SiteAccess usage)
        parents = request.PARENTS
        parents.pop() # I don't belong there

        if len(name) > 1:
            if name.startswith("//"): # we're doing alias tricks (new as of VHF)
                obj = self
                for id in name[2:].split('/'):
                    obj=obj._getOb(id) # if there's an error, then we have an error in the program logic, so let it shine
                ret=obj.getTraversalResponse()
                if IHostOverride.ISimpleAliasHostOverride.isImplementedBy(ret):
                    request.setAliasFolder(ret.getId())
                    return ret.__of__(parents[-1]) # just a single alias
                # otherwise, ret should be a list, the first of which
                # should be linked by aquisition to the Zope root and
                # the rest linked to the previous one down the chain.
                # Currently not using this traversal feature (more
                # discussion can be found on this in the Alias.py
                # getTraversalResponse method definition)
                return ret
            else: request.setVirtualRoot(name[1:].split('/'))
        else:
            request.setVirtualRoot([])
        return parents.pop() # He'll get put back on

    security.declareProtected( 'View management screens', 'addToContainer' )
    def addToContainer(self, container):
        """the magic begins in manage_afterAdd"""
        container._setObject(self.getId(), self)
        self.manage_afterAdd(self, container)

    security.declareProtected( 'View management screens', 'manage_addToContainer' )
    def manage_addToContainer(self, container, nextURL=''):
        """The Zope management wrapper around addToContainer: the main point is
        to eventually get to the afterAdd hook"""
        self.addToContainer(container)
        if nextURL:
            msg="""  It is currently active, ready to process new
            requests if the front-end server has been set up
            correctly."""
            if not self.getHookStatus(container):
                msg="""  It is not currently active, probably because
                you are already using the Virtual Host Monster
                product.  Use the evaluate tab of the Virtual Host
                Folder to switch smoothly over to this new VHF, and to
                switch back if you decide the VHM suits your needs
                better."""
            return MessageDialog(title='Item Added',
              message='This object now has a %s.%s' % (self.meta_type, msg),
              action=nextURL)

    security.declareProtected( 'View management screens', 'unhook' )
    def unhook(self, container):
        """removes self from BeforeTraverse"""
        BeforeTraverse.unregisterBeforeTraverse(container, self.meta_type)

    security.declareProtected( 'View management screens', 'hook' )
    def hook(self, container):
        """hooks self up to BeforeTraverse"""
        id = self.id
        if callable(id): id = id()
        # We want the original object, not stuff in between
        container = container.this()
        hook = BeforeTraverse.NameCaller(id)
        BeforeTraverse.registerBeforeTraverse(container, hook,
                                              self.meta_type,
                                              self.priority)

    security.declareProtected( 'View management screens', 'unhookAndHookVHM' )
    def unhookAndHookVHM(self, container):
        """unhooks self and tries to hook up virtual host monster, if present, to BeforeTraverse"""
        if self.getHookStatus(container):
            self.unhook(container)
        vhm=container.objectItems( "Virtual Host Monster" )
        if vhm:
            vhm=vhm[0][1]
            if not BeforeTraverse.queryBeforeTraverse(container.this(), vhm.meta_type):
                id=vhm.id
                if callable(id): id=id()
                container=container.this()
                hook=BeforeTraverse.NameCaller(id)
                BeforeTraverse.registerBeforeTraverse(container, hook,
                                                      vhm.meta_type,
                                                      vhm.priority)

    security.declareProtected( 'View management screens', 'getHookStatus' )
    def getHookStatus(self, container=None):
        """returns whether this object is currently hooked up to BeforeTraverse"""
        id = self.id
        if callable(id): id = id()
        if not container:
            container=aq_parent(self)
        #took this out because we want an exception raised--it indicates a bug in the program logic
        #if not container or not getattr(container, id) == self:
        #    raise SyntaxError, ""
        return BeforeTraverse.queryBeforeTraverse(container.this(), self.meta_type)

    security.declareProtected( 'View management screens', 'hookAndUnhookVHM' )
    def hookAndUnhookVHM(self, container):
        """hooks up the VHF with BeforeTraverse, and unhooks the VHM if it is
        present."""
        vhm=container.objectItems( "Virtual Host Monster" )
        if vhm:
            vhm=vhm[0][1]
            if BeforeTraverse.queryBeforeTraverse(container.this(), vhm.meta_type):
                BeforeTraverse.unregisterBeforeTraverse(container, vhm.meta_type)
        if not self.getHookStatus(container):
            self.hook(container)

    security.declareProtected( 'View management screens', 'manage_beforeDelete' )
    def manage_beforeDelete(self, item, container):
        """takes care of unhooking the traversal call; tries to reinstate VHM if present"""
        if item is self:
            self.unhookAndHookVHM(container)
            # BeforeTraverse.unregisterBeforeTraverse(container, self.meta_type)

    security.declareProtected( 'View management screens', 'manage_afterAdd' )
    def manage_afterAdd(self, item, container):
        """This hooks up the __call__ function as a way to pre-process the
        traversal of the path so it can modify things"""
        if not container.objectItems( "Virtual Host Monster" ):
            self.hook(container)

    # security.declarePrivate( '_setId' ) # currently unnecessary
    security.declarePrivate('oldSetId')
    oldSetId=Folder.inheritedAttribute('_setId')
    def _setId(self, id):
        """necessary for id changes; I better watch out to use this when
        necessary"""
        id = str(id)
        if id != self.getId() and self.getHookStatus():
            self.unhook(aq_parent(self))
            oldSetId(id)
            self.hook(aq_parent(self))
        else: oldSetId(id)
    setId=_setId


################################################################################
############ API

    # specify that VHF is responsible for providing breadcrumbs presumably in __call__; specify that host (or override) is responsible for setting VirtualRootPhysicalPath, CANONICAL_VIRTUAL_HOST_ROOT_PATH


    security.declarePublic( 'getCanonicalVirtualRootURL' )
    def getCanonicalVirtualRootURL(self, REQUEST=None):
        """Get the canonical virtual host root URL (string)"""
        if not REQUEST and hasattr(self, "REQUEST"):
            REQUEST=self.REQUEST
        if REQUEST and REQUEST.has_key("CANONICAL_VIRTUAL_HOST_ROOT_PATH"):
            return REQUEST["CANONICAL_VIRTUAL_HOST_ROOT_PATH"]
        return None

    security.declarePublic( 'getCanonicalVirtualRootObject' )
    def getCanonicalVirtualRootObject(self, REQUEST=None):
        """Get the canonical virtual host root object"""
        try:
            return self.unrestrictedTraverse(self.getCanonicalVirtualRootURL(REQUEST))
        except:
            return None

    security.declarePublic( 'getActualVirtualRootURL' )
    def getActualVirtualRootURL(self, REQUEST=None):
        """Get the actual virtual host root URL (string); it will be
        the same as the above, unless overwritten by a HostOverride
        object such as an alias."""
        if not REQUEST and hasattr(self, "REQUEST"):
            REQUEST=self.REQUEST
        if REQUEST and REQUEST.has_key("VirtualRootPhysicalPath"):
            return REQUEST["VirtualRootPhysicalPath"]
        return None

    security.declarePublic( 'getActualVirtualRootObject' )
    def getActualVirtualRootObject(self, REQUEST=None):
        """Get the actual virtual host root object, if any"""
        try:
            return self.unrestrictedTraverse(self.getActualVirtualRootURL(REQUEST))
        except:
            return None

    def getCurrentVirtualHostIdsURL(self, REQUEST=None):
        """Get the actual virtual host list of ids (host and aliases, if
        any)."""
        if not REQUEST and hasattr(self, "REQUEST"):
            REQUEST=self.REQUEST
        if REQUEST and REQUEST.has_key("VIRTUAL_HOST_IDS"):
            return REQUEST["VIRTUAL_HOST_IDS"]
        return None

    def getOriginalRequestURL(self, REQUEST=None):
        """Get the literal slash-separated string, minus host port and protocol, used in this request"""
        if not REQUEST and hasattr(self, "REQUEST"):
            REQUEST=self.REQUEST
        if REQUEST and REQUEST.has_key("VIRTUAL_HOST_BREADCRUMBS"):
            return REQUEST["VIRTUAL_HOST_BREADCRUMBS"]
        return None

    def getCleanOriginalRequestList(self, REQUEST=None, rootName="/", includeQueryString=0):
        """Returns a list representing a cleaned version of the
        original request combined with breadcrumb links for each
        element.  index_html is removed, and a final element with a
        query string hides the query string in the path element but
        includes it in the URL.  Intended to save time with
        breadcrumbs.  Format is this, for instance:
        [
            {"id":"","url":"/"},
            {"id":"firstPathElement","url":"/firstPathElement"},
            {"id":"lastPathElement","url":"/firstPathElement/lastPathElement?queryString=Hidden"}
        ]"""
        return utils.pathToBreadcrumbs(self.getOriginalRequestURL(REQUEST).split("/"), rootName, includeQueryString)


    security.declarePublic( 'getCurrentVirtualHostObject' )
    def getCurrentVirtualHostObject(self, REQUEST=None):
        """returns the current virtual host object, or None. (Note
        that in the current implementation each host can have a
        property sheet, which can be modified using standard Zope
        property sheet interfaces after the current host is obtained
        via this method.  I deemed an overlay interface to modify the
        properties via the virtual host folder to be
        counter-productive, for now.)"""
        if not REQUEST and hasattr(self, "REQUEST"):
            REQUEST=self.REQUEST
        hostid=self.getCurrentVirtualHostId(REQUEST)
        if hostid:
            return self._getOb(hostid)
        return None

    security.declarePublic( 'getVirtualHostStatus' )
    def getVirtualHostStatus(self, REQUEST=None):
        """returns 1 if a virtual host is currently in use, otherwise 0"""
        if not REQUEST and hasattr(self, "REQUEST"):
            REQUEST=self.REQUEST
        return REQUEST and REQUEST.has_key("VIRTUAL_HOST_IDS") and REQUEST["VIRTUAL_HOST_IDS"] and 1

    security.declarePublic( 'getCurrentVirtualHostId' )
    def getCurrentVirtualHostId(self, REQUEST=None):
        """Just returns the string id of the currently active host, or None"""
        if not REQUEST and hasattr(self, "REQUEST"):
            REQUEST=self.REQUEST
        if REQUEST and REQUEST.has_key("VIRTUAL_HOST_IDS"):
            return REQUEST["VIRTUAL_HOST_IDS"].split("/")[0]
        return None

#    def virtualContextTraverse(self, url, REQUEST=None, unrestricted=0): # is this needed?
#        """given a URL, tries to treat it as a request for an object from the
#        current canonical virtual root."""
#        rt=self.getCanonicalVirtualRootURL(REQUEST)


################################################################################

manage_addVirtualHostFolderForm=DTMLFile('dtml/addVirtualHostFolder', globals())

def manage_addVirtualHostFolder(dispatcher, id, REQUEST=None, title="", **ignored):
    """Constructor for the Virtual Host Folder"""
    container=dispatcher.Destination()
    rootObj=container.getPhysicalRoot()
    if container==rootObj and not rootObj.objectItems( "Virtual Host Folder" ):
        vhf = VirtualHostFolder()
        vhf.oldSetId(str(id))
        if title: vhf.title=title
        if REQUEST:
            return vhf.manage_addToContainer(container,
                                        '%s/manage_main' % REQUEST['URL1'])
        else:
            vhf.addToContainer(container)
            return 1
    if REQUEST:
        return MessageDialog(title='ERROR: Item NOT Added',
                message="""ERROR: A Virtual Host Folder can only be added in the Zope
                root folder, and only if a Virtual Host Folder does not
                exist there already.""",
                action='%s/manage_main' % REQUEST['URL1'])
    else:
        return 0

constructors = (
  ('manage_addVirtualHostFolderForm', DTMLFile('www/addVirtualHostFolder', globals())),
  ('manage_addVirtualHostFolder', manage_addVirtualHostFolder),
)


InitializeClass(VirtualHostFolder)
